from __future__ import annotations

import logging
from dataclasses import dataclass, field
from typing import Any, Iterator, OrderedDict, Type

from src.stairlight.source.config_key import MapKey

logger = logging.getLogger()


class ConfigAttributeNotFoundException(Exception):
    def __init__(self, msg: str) -> None:
        self.msg = msg

    def __str__(self) -> str:
        return self.msg


@dataclass
class StairlightConfigInclude:
    TemplateSourceType: str


@dataclass
class StairlightConfigExclude:
    TemplateSourceType: str | None = None
    Regex: str | None = None


@dataclass
class StairlightConfigSettings:
    MappingPrefix: str | None = None


@dataclass
class StairlightConfig:
    Include: list[dict[str, Any]] = field(default_factory=list)
    Exclude: list[dict[str, Any]] = field(default_factory=list)
    Settings: OrderedDict = field(default_factory=OrderedDict)

    @staticmethod
    def select_config_include(source_type: str) -> Type[StairlightConfigInclude]:
        """Select a data class of include section by source type

        Args:
            source_type (str): Source type

        Returns:
            Type[StairlightConfigInclude]: Include section
        """
        from src.stairlight.source.template import TemplateSourceType

        config_include: Type[StairlightConfigInclude] = StairlightConfigInclude

        # Avoid to occur circular imports
        if source_type == TemplateSourceType.FILE.value:
            from src.stairlight.source.file.config import StairlightConfigIncludeFile

            config_include = StairlightConfigIncludeFile
        elif source_type == TemplateSourceType.GCS.value:
            from src.stairlight.source.gcs.config import StairlightConfigIncludeGcs

            config_include = StairlightConfigIncludeGcs
        elif source_type == TemplateSourceType.REDASH.value:
            from src.stairlight.source.redash.config import (
                StairlightConfigIncludeRedash,
            )

            config_include = StairlightConfigIncludeRedash
        elif source_type == TemplateSourceType.DBT.value:
            from src.stairlight.source.dbt.config import StairlightConfigIncludeDbt

            config_include = StairlightConfigIncludeDbt
        elif source_type == TemplateSourceType.S3.value:
            from src.stairlight.source.s3.config import StairlightConfigIncludeS3

            config_include = StairlightConfigIncludeS3
        return config_include

    def get_include(self) -> Iterator[StairlightConfigInclude]:
        """Get attributes of a include section

        Yields:
            Iterator[StairlightConfigInclude]: Include section
        """
        for _include in self.Include:
            config = self.select_config_include(
                source_type=str(_include.get(MapKey.TEMPLATE_SOURCE_TYPE))
            )
            yield config(**_include)

    def get_exclude(self) -> Iterator[StairlightConfigExclude]:
        """Get attributes of a exclude section

        Yields:
            Iterator[StairlightConfigExclude]: Exclude section
        """
        for _exclude in self.Exclude:
            yield StairlightConfigExclude(**_exclude)


@dataclass
class MappingConfigGlobal:
    Parameters: dict[str, Any] | None = None


@dataclass
class MappingConfigMappingTable:
    TableName: str
    IgnoreParameters: list[str] | None = None
    Parameters: OrderedDict | None = None
    Labels: dict[str, Any] | None = None


@dataclass
class MappingConfigMapping:
    TemplateSourceType: str
    Tables: list[OrderedDict] = field(default_factory=list)

    def get_table(self) -> Iterator[MappingConfigMappingTable]:
        for _table in self.Tables:
            yield MappingConfigMappingTable(**_table)


@dataclass
class MappingConfigExtraLabels:
    TableName: str | None = None
    Labels: dict[str, Any] = field(default_factory=dict)


@dataclass
class MappingConfig:
    Global: OrderedDict | None = None
    Mapping: list[OrderedDict] = field(default_factory=list)
    ExtraLabels: list[dict[str, Any]] | None = None
    Metadata: list[dict[str, Any]] | None = None  # Deprecated

    def get_global(self) -> MappingConfigGlobal:
        """Get global section

        Returns:
            MappingConfigGlobal: Global section
        """
        if self.Global:
            mapping_config_global = MappingConfigGlobal(**self.Global)
        else:
            mapping_config_global = MappingConfigGlobal()
        return mapping_config_global

    def get_mapping(self) -> Iterator[MappingConfigMapping]:
        """Get mapping section

        Yields:
            Iterator[MappingConfigMapping]: Mapping section
        """
        for _mapping in self.Mapping:
            mapping_config = self.select_mapping_config(
                source_type=str(_mapping.get(MapKey.TEMPLATE_SOURCE_TYPE))
            )
            yield mapping_config(**_mapping)

    def get_extra_labels(self) -> Iterator[MappingConfigExtraLabels]:
        """Get extra labels section

        Yields:
            Iterator[MappingConfigExtraLabels]: Extra labels section
        """
        if not self.ExtraLabels:
            return
        for extra_label in self.ExtraLabels:
            yield MappingConfigExtraLabels(**extra_label)

        # Deprecated
        if not self.Metadata:
            return
        for metadata in self.Metadata:
            yield MappingConfigExtraLabels(**metadata)

    @staticmethod
    def select_mapping_config(source_type: str) -> Type[MappingConfigMapping]:
        """Select a mapping data class from source type

        Args:
            source_type (str): Source type

        Returns:
            Type[MappingConfigMapping]: Mapping section
        """
        from src.stairlight.source.template import TemplateSourceType

        mapping_config: Type[MappingConfigMapping] = MappingConfigMapping

        # Avoid to occur circular imports
        if source_type == TemplateSourceType.FILE.value:
            from src.stairlight.source.file.config import MappingConfigMappingFile

            mapping_config = MappingConfigMappingFile
        elif source_type == TemplateSourceType.GCS.value:
            from src.stairlight.source.gcs.config import MappingConfigMappingGcs

            mapping_config = MappingConfigMappingGcs
        elif source_type == TemplateSourceType.REDASH.value:
            from src.stairlight.source.redash.config import MappingConfigMappingRedash

            mapping_config = MappingConfigMappingRedash
        elif source_type == TemplateSourceType.DBT.value:
            from src.stairlight.source.dbt.config import MappingConfigMappingDbt

            mapping_config = MappingConfigMappingDbt
        elif source_type == TemplateSourceType.S3.value:
            from src.stairlight.source.s3.config import MappingConfigMappingS3

            mapping_config = MappingConfigMappingS3
        return mapping_config



from src.stairlight.stairlight import (
    ResponseType,
    SearchDirection,
    StairLight,
    is_cyclic,
)

__version__ = "0.7.1"
__all__ = ["StairLight", "ResponseType", "SearchDirection", "is_cyclic", "__version__"]



from __future__ import annotations

import re
from typing import Iterator

from src.stairlight.source.config import MapKey


class Query:
    """SQL query"""

    def __init__(self, query_str: str, default_table_prefix: str = None) -> None:
        """SQL query

        Args:
            query_str (str): Query statement.
            default_table_prefix (str, optional):
                If project or dataset that configured table have are omitted,
                it will be complement this prefix. Defaults to None.
        """
        self.query_str = query_str
        self.default_table_prefix = default_table_prefix

    def detect_upstairs_attributes(self) -> Iterator[dict]:
        """Parse a query statement and detect upstream table attributes

        Yields:
            Iterator[dict]: upstream table attributes
        """
        upstairs_tables = self.parse_and_get_upstairs_tables()

        for upstairs_table in upstairs_tables:
            line_indexes = [
                i
                for i, line in enumerate(self.query_str.splitlines())
                if upstairs_table in line
                and "--" not in line.split(upstairs_table)[0]  # exclude comments
            ]

            for line_index in line_indexes:
                if self.default_table_prefix:
                    table_name = solve_table_prefix(
                        table=upstairs_table,
                        default_table_prefix=self.default_table_prefix,
                    )
                else:
                    table_name = upstairs_table
                yield {
                    MapKey.TABLE_NAME: table_name.replace("`", ""),  # for BigQuery
                    MapKey.LINE_NUMBER: line_index + 1,
                    MapKey.LINE_STRING: self.query_str.splitlines()[line_index],
                }

    def parse_and_get_upstairs_tables(self) -> list[str]:
        """Parse query and get upstairs tables

        Returns:
            list[str]: A list of upstairs tables
        """
        # Remove comments
        comments_pattern = r"\-\-.*\n"
        query_str = re.sub(comments_pattern, "", self.query_str)

        # Get Common-Table-Expressions(CTE) from query string
        cte_pattern = r"(?:with|,)\s*(\w+)\s+as\s*"
        cte_alias: list[str] = re.findall(cte_pattern, query_str, re.IGNORECASE)

        # Search a boundary line number that a main query starts
        boundary_num: int = 0
        main_pattern = r"\)[;\s]*select" if any(cte_alias) else r"select"
        main_search_result = re.search(main_pattern, query_str, re.IGNORECASE)
        if main_search_result:
            boundary_num = main_search_result.start()

        # Split the query to a main query and CTEs
        query_group = {}
        query_group["main"] = query_str[boundary_num:].strip()
        query_group["cte"] = query_str[:boundary_num].strip()

        table_pattern = r"\s(?:from|join)\s+([`.\-\w]+)"
        main_tables_with_alias: list[str] = re.findall(
            table_pattern, query_group["main"], re.IGNORECASE
        )

        # Exclude Google BigQuery EXTRACT function
        bq_extract_pattern = r"(?:EXTRACT\(.+ FROM)\s+([`.\-\w]+)"
        bq_extracts: list[str] = re.findall(
            bq_extract_pattern, query_group["main"], re.IGNORECASE
        )

        main_tables = [
            table
            for table in main_tables_with_alias
            if table not in cte_alias
            and table not in bq_extracts
            and table.upper() != "UNNEST"
        ]

        # Exclude table alias from CTEs
        cte_tables_with_alias: list[str] = re.findall(
            table_pattern, query_group["cte"], re.IGNORECASE
        )
        cte_tables = [
            cte_table
            for cte_table in cte_tables_with_alias
            if cte_table not in cte_alias
        ]

        return sorted(set(main_tables + cte_tables))


def solve_table_prefix(table: str, default_table_prefix: str) -> str:
    """Solve table name prefix

    Args:
        table (str): Table name
        default_table_prefix (str):
            If project or dataset that configured table have are omitted,
            it will be complement this prefix. Defaults to None.

    Returns:
        str: Solved table name.
    """
    solved_name = table
    if table.count(".") <= 1:
        solved_name = default_table_prefix + "." + solved_name
    return solved_name



from __future__ import annotations

import json
import os
from importlib.util import find_spec
from logging import getLogger
from pathlib import Path
from typing import Any, OrderedDict, Type

from src.stairlight.source.config import MappingConfigMapping
from src.stairlight.source.config_key import GCS_URI_SCHEME, S3_URI_SCHEME
from src.stairlight.source.dbt.config import MappingConfigMappingDbt
from src.stairlight.source.file.config import MappingConfigMappingFile
from src.stairlight.source.gcs.config import MappingConfigMappingGcs
from src.stairlight.source.redash.config import MappingConfigMappingRedash
from src.stairlight.source.s3.config import MappingConfigMappingS3
from src.stairlight.source.template import Template, TemplateSource, TemplateSourceType

logger = getLogger(__name__)


def get_template_source_class(template_source_type: str) -> Type[TemplateSource]:
    """Get a template source class for dynamic importing

    Args:
        template_source_type (str): Template source type

    Returns:
        TemplateSource: Template source class
    """
    template_source: Type[TemplateSource]
    if template_source_type == TemplateSourceType.FILE.value:
        if not find_spec("FileTemplateSource"):
            from src.stairlight.source.file.template import FileTemplateSource
        template_source = FileTemplateSource
    elif template_source_type == TemplateSourceType.GCS.value:
        if not find_spec("GcsTemplateSource"):
            from src.stairlight.source.gcs.template import GcsTemplateSource
        template_source = GcsTemplateSource
    elif template_source_type == TemplateSourceType.REDASH.value:
        if not find_spec("RedashTemplateSource"):
            from src.stairlight.source.redash.template import RedashTemplateSource
        template_source = RedashTemplateSource
    elif template_source_type == TemplateSourceType.DBT.value:
        if not find_spec("DbtTemplateSource"):
            from src.stairlight.source.dbt.template import DbtTemplateSource
        template_source = DbtTemplateSource
    elif template_source_type == TemplateSourceType.S3.value:
        if not find_spec("S3TemplateSource"):
            from src.stairlight.source.s3.template import S3TemplateSource
        template_source = S3TemplateSource
    return template_source


def get_default_table_name(template: Template) -> str:
    """Get a default table name

    Args:
        template (Template): Query template

    Returns:
        str: Default table name
    """
    default_table_name: str = ""
    if template.source_type == TemplateSourceType.REDASH:
        default_table_name = template.uri
    else:
        default_table_name = Path(template.key).stem
    return default_table_name


def collect_mapping_attributes(
    template: Template,
    tables: list[OrderedDict[str, Any]],
) -> MappingConfigMapping:
    """Collect attributes from a mapping section

    Args:
        template (Template): Query template
        tables (list[OrderedDict[str, Any]]): Tables

    Returns:
        MappingConfigMapping: Attributes of a mapping section
    """
    mapping: MappingConfigMapping

    if template.source_type == TemplateSourceType.FILE:
        mapping = MappingConfigMappingFile(
            FileSuffix=template.key,
            Tables=tables,
        )
    elif template.source_type == TemplateSourceType.GCS:
        mapping = MappingConfigMappingGcs(
            Uri=template.uri,
            Tables=tables,
        )
    elif template.source_type == TemplateSourceType.REDASH:
        mapping = MappingConfigMappingRedash(
            QueryId=template.query_id,
            DataSourceName=template.data_source_name,
            Tables=tables,
        )
    elif template.source_type == TemplateSourceType.DBT:
        mapping = MappingConfigMappingDbt(
            ProjectName=template.project_name,
            FileSuffix=template.key,
            Tables=tables,
        )
    elif template.source_type == TemplateSourceType.S3:
        mapping = MappingConfigMappingS3(
            Uri=template.uri,
            Tables=tables,
        )

    return mapping


class SaveMapController:
    def __init__(self, save_file: str, mapped: dict[str, Any]) -> None:
        self.save_file = save_file
        self._mapped = mapped

    def save(self) -> None:
        """Save mapped results"""
        if self.save_file.startswith(GCS_URI_SCHEME):
            self._save_map_gcs()
        elif self.save_file.startswith(S3_URI_SCHEME):
            self._save_map_s3()
        else:
            self._save_map_file()

    def _save_map_file(self) -> None:
        """Save mapped results to file system"""
        with open(self.save_file, "w") as f:
            json.dump(self._mapped, f, indent=2)

    def _save_map_gcs(self) -> None:
        """Save mapped results to Google Cloud Storage"""
        from google.cloud.storage import Blob

        from src.stairlight.source.gcs.map import get_gcs_blob

        blob: Blob = get_gcs_blob(uri=self.save_file)
        blob.upload_from_string(
            data=json.dumps(obj=self._mapped, indent=2),
            content_type="application/json",
        )

    def _save_map_s3(self) -> None:
        """Save mapped results to Amazon S3"""
        from mypy_boto3_s3.service_resource import Object

        from src.stairlight.source.s3.map import get_s3_object

        _object: Object = get_s3_object(uri=self.save_file)
        _ = _object.put(Body=json.dumps(obj=self._mapped, indent=2))


class LoadMapController:
    def __init__(self, load_file: str) -> None:
        self.load_file = load_file

    def load(self) -> dict:
        """Load mapped results

        Returns:
            dict: Loaded map
        """
        loaded_map = {}
        if self.load_file.startswith(GCS_URI_SCHEME):
            loaded_map = self._load_map_gcs()
        elif self.load_file.startswith(S3_URI_SCHEME):
            loaded_map = self._load_map_s3()
        else:
            loaded_map = self._load_map_file()
        return loaded_map

    def _load_map_file(self) -> dict:
        """Load mapped results from file system"""
        if not os.path.exists(self.load_file):
            logger.error(f"{self.load_file} is not found.")
            exit()
        with open(self.load_file) as f:
            return json.load(f)

    def _load_map_gcs(self) -> dict:
        """Load mapped results from Google Cloud Storage"""
        from google.cloud.storage import Blob

        from src.stairlight.source.gcs.map import get_gcs_blob

        blob: Blob = get_gcs_blob(uri=self.load_file)
        if not blob.exists():
            logger.error(f"{self.load_file} is not found.")
            exit()
        return json.loads(blob.download_as_string())

    def _load_map_s3(self) -> dict:
        """Load mapped results from Amazon S3"""
        from botocore.response import StreamingBody
        from mypy_boto3_s3.service_resource import Object
        from mypy_boto3_s3.type_defs import GetObjectOutputTypeDef

        from src.stairlight.source.s3.map import get_s3_object

        _object: Object = get_s3_object(uri=self.load_file)
        object_output: GetObjectOutputTypeDef = _object.get()
        body: StreamingBody = object_output["Body"]
        if not body:
            logger.error(f"{self.load_file} is not found.")
            exit()
        return json.loads(body.read().decode("utf-8"))



from __future__ import annotations

from dataclasses import dataclass, field
from typing import Any, OrderedDict

from src.stairlight.source.config import MappingConfigMapping, StairlightConfigInclude
from src.stairlight.source.template import TemplateSourceType as source_type


@dataclass
class StairlightConfigIncludeFile(StairlightConfigInclude):
    TemplateSourceType: str = source_type.FILE.value
    FileSystemPath: str | None = None
    Regex: str | None = None
    DefaultTablePrefix: str | None = None


@dataclass
class MappingConfigMappingFile(MappingConfigMapping):
    TemplateSourceType: str = source_type.FILE.value
    FileSuffix: str | None = None
    Tables: list[OrderedDict[str, Any]] = field(default_factory=list)



from __future__ import annotations

import glob
import os
import pathlib
import re
import shlex
import subprocess
from typing import Any, Iterator

import yaml

from src.stairlight.source.config import MappingConfig, StairlightConfig
from src.stairlight.source.config_key import DbtProjectKey
from src.stairlight.source.dbt.config import StairlightConfigIncludeDbt
from src.stairlight.source.template import Template, TemplateSource, TemplateSourceType


class DbtTemplate(Template):
    def __init__(
        self,
        mapping_config: MappingConfig | None,
        key: str,
        project_name: str,
    ):
        super().__init__(
            mapping_config=mapping_config,
            key=key,
            source_type=TemplateSourceType.DBT,
        )
        self.uri = self.get_uri()
        self.project_name = project_name

    def get_uri(self) -> str:
        """Get uri from a key

        Returns:
            str: uri
        """
        return str(pathlib.Path(self.key).resolve())

    def get_template_str(self) -> str:
        """Get template string that read from a key

        Returns:
            str: Template string
        """
        with open(self.key) as f:
            return f.read()

    def render(
        self, params: dict[str, Any] = None, ignore_params: list[str] = None
    ) -> str:
        """Render a query statement from a jinja template

        Args:
            params (dict[str, Any]): Jinja parameters
            ignore_params (list[str]): Ignore parameters. Defaults to None.

        Returns:
            str: Query statement
        """
        return self.get_template_str()


class DbtTemplateSource(TemplateSource):
    DBT_PROJECT_YAML = "dbt_project.yml"
    REGEX_SCHEMA_TEST_FILE = re.compile(r".*/schema.yml/.*\.sql$")

    def __init__(
        self,
        stairlight_config: StairlightConfig,
        mapping_config: MappingConfig,
        include: StairlightConfigIncludeDbt,
    ) -> None:
        super().__init__(
            stairlight_config=stairlight_config,
            mapping_config=mapping_config,
        )
        self._include = include

    def search_templates(self) -> Iterator[Template]:
        """Search query template files

        Yields:
            Iterator[Template]: Attributes of query template files
        """
        if not self._include:
            return None

        project_dir: str | None = self._include.ProjectDir
        profiles_dir: str | None = self._include.ProfilesDir
        if not project_dir or not profiles_dir:
            return None

        dbt_project_config: dict[str, Any] = self.read_dbt_project_yml(
            project_dir=project_dir
        )
        _ = self.execute_dbt_compile(
            project_dir=project_dir,
            profiles_dir=profiles_dir,
            profile=dbt_project_config.get(DbtProjectKey.PROFILE),
            target=self._include.Target,
            vars=self._include.Vars,
        )

        target_path = dbt_project_config[DbtProjectKey.TARGET_PATH]
        project_name = dbt_project_config[DbtProjectKey.PROJECT_NAME]
        model_paths = dbt_project_config[DbtProjectKey.MODEL_PATHS]

        for model_path in model_paths:
            dbt_model_path_str = self.concat_dbt_model_path_str(
                project_dir=project_dir,
                target_path=target_path,
                project_name=project_name,
                model_path=model_path,
            )
            dbt_model_path = pathlib.Path(dbt_model_path_str)
            for p in dbt_model_path.glob("**/*"):
                if self.is_skipped(p=p):
                    self.logger.debug(f"{str(p)} is skipped.")
                    continue

                yield DbtTemplate(
                    mapping_config=self._mapping_config,
                    key=str(p),
                    project_name=project_name,
                )

    def is_skipped(self, p: pathlib.Path) -> bool:
        """Check the target path is skipped or not

        Args:
            p (pathlib.Path): Path

        Returns:
            bool: Is skipped or not
        """
        is_matched = False
        if self.REGEX_SCHEMA_TEST_FILE.fullmatch(str(p)):
            is_matched = True
        return (
            p.is_dir()
            or is_matched
            or self.is_excluded(
                source_type=TemplateSourceType(self._include.TemplateSourceType),
                key=str(p),
            )
        )

    def read_dbt_project_yml(self, project_dir: str) -> dict:
        """Read dbt_project.yml

        Args:
            project_dir (str): dbt project directory

        Returns:
            dict: dbt project settings
        """
        dbt_project_pattern = re.compile(f"^{project_dir}/{self.DBT_PROJECT_YAML}$")
        return self.read_yml(dir=project_dir, re_pattern=dbt_project_pattern)

    def read_yml(self, dir: str, re_pattern: re.Pattern) -> dict:
        """Read DBT settings from a yml file

        Args:
            dir (str): Directory
            re_pattern (re.Pattern): Regular expression pattern

        Returns:
            dict: Results
        """
        files = [
            obj
            for obj in glob.glob(f"{dir}/**", recursive=False)
            if re_pattern.fullmatch(obj)
        ]
        if not files:
            self.logger.error(f"Not found: {re_pattern} in {os.getcwd()}")
            exit()
        with open(files[0]) as file:
            return yaml.safe_load(file)

    @staticmethod
    def concat_dbt_model_path_str(
        project_dir: str,
        target_path: str,
        project_name: str,
        model_path: pathlib.Path,
    ) -> str:
        """Return DBT model path

        Args:
            project_dir (str): DBT project directory
            target_path (str): DBT target
            project_name (str): DBT project
            model_path (pathlib.Path): DBT model path

        Returns:
            str: DBT model path
        """
        return (
            f"{project_dir}/"
            f"{target_path}/"
            "compiled/"
            f"{project_name}/"
            f"{model_path}/"
        )

    @staticmethod
    def build_dbt_compile_command(
        project_dir: str,
        profiles_dir: str,
        profile: str = None,
        target: str = None,
        vars: dict[str, Any] = None,
    ) -> str:
        """Build a DBT compile command

        Args:
            project_dir (str): DBT project directory
            profiles_dir (str): DBT profile directory
            profile (str, optional): DBT profile. Defaults to None.
            target (str, optional): DBT target. Defaults to None.
            vars (dict[str, Any], optional): DBT variables. Defaults to None.

        Returns:
            str: DBT compile command
        """
        command = (
            "dbt compile"
            f" --project-dir {project_dir}"
            f" --profiles-dir {profiles_dir}"
        )
        if profile:
            command += f" --profile {profile}"
        if target:
            command += f" --target {target}"
        if vars:
            command += f" --vars '{vars}'"
        return command

    def execute_dbt_compile(
        self,
        project_dir: str,
        profiles_dir: str,
        profile: str = None,
        target: str = None,
        vars: dict[str, Any] = None,
    ) -> int:
        """Execute dbt compile

        Args:
            project_dir (str): DBT project directory
            profiles_dir (str): DBT profile directory
            profile (str, optional): DBT profile. Defaults to None.
            target (str, optional): DBT target. Defaults to None.
            vars (dict[str, Any], optional): DBT variables. Defaults to None.

        Returns:
            int: Return code
        """
        command = self.build_dbt_compile_command(
            project_dir=project_dir,
            profiles_dir=profiles_dir,
            profile=profile,
            target=target,
            vars=vars,
        )
        proc = subprocess.run(
            args=shlex.split(command),
            shell=False,
            check=True,
            stdout=subprocess.DEVNULL,
        )
        return proc.returncode



from __future__ import annotations

import glob
import logging
import re
from collections import OrderedDict
from dataclasses import asdict
from datetime import datetime, timezone
from typing import Any

import yaml

from src.stairlight.source.config import (
    MappingConfig,
    MappingConfigMapping,
    MappingConfigMappingTable,
    StairlightConfig,
    StairlightConfigExclude,
    StairlightConfigSettings,
)
from src.stairlight.source.config_key import MapKey
from src.stairlight.source.controller import (
    collect_mapping_attributes,
    get_default_table_name,
)
from src.stairlight.source.dbt.config import StairlightConfigIncludeDbt
from src.stairlight.source.file.config import StairlightConfigIncludeFile
from src.stairlight.source.gcs.config import StairlightConfigIncludeGcs
from src.stairlight.source.redash.config import StairlightConfigIncludeRedash
from src.stairlight.source.s3.config import StairlightConfigIncludeS3
from src.stairlight.source.template import Template

logger = logging.getLogger()

STAIRLIGHT_CONFIG_PREFIX_DEFAULT = "stairlight"
MAPPING_CONFIG_PREFIX_DEFAULT = "mapping"


class Configurator:
    def __init__(self, dir: str) -> None:
        """Configuration class

        Args:
            dir (str): A directory that Configuration files exists.
        """
        self.dir = dir

    def read_stairlight(
        self, prefix: str = STAIRLIGHT_CONFIG_PREFIX_DEFAULT
    ) -> StairlightConfig:
        """Read stairlight configurations from yaml

        Args:
            prefix (str, optional):
                Prefix of the configuration file name.
                Defaults to STAIRLIGHT_CONFIG_PREFIX_DEFAULT.

        Returns:
            StairlightConfig: Stairlight configurations
        """
        config = self.read(prefix=prefix)
        return StairlightConfig(**config)

    def read_mapping(
        self, prefix: str = MAPPING_CONFIG_PREFIX_DEFAULT
    ) -> MappingConfig:
        """Read mapping configurations from yaml

        Args:
            prefix (str, optional):
                Prefix of the configuration file name.
                Defaults to MAPPING_CONFIG_PREFIX_DEFAULT.

        Returns:
            MappingConfig: Mapping configurations
        """
        config = self.read(prefix=prefix)
        return MappingConfig(**config)

    def read(self, prefix: str) -> dict[str, Any]:
        """Read a configuration file

        Args:
            prefix (str): Prefix of configuration file name.

        Returns:
            dict: configurations
        """
        config: dict[str, Any] = {}
        pattern = f"^{self.dir}/{prefix}.ya?ml$"
        config_file = [
            p
            for p in glob.glob(f"{self.dir}/**", recursive=False)
            if re.fullmatch(pattern, p)
        ]
        if config_file:
            with open(config_file[0]) as file:
                config = yaml.safe_load(file)
        return config

    def create_stairlight_file(
        self, prefix: str = STAIRLIGHT_CONFIG_PREFIX_DEFAULT
    ) -> str:
        """Create a Stairlight template file

        Args:
            prefix (str, optional):
                Prefix of the configuration file name.
                Defaults to STAIRLIGHT_CONFIG_PREFIX.

        Returns:
            str: Created file name
        """
        template_file_name = f"{self.dir}/{prefix}.yaml"
        with open(template_file_name, "w") as f:
            yaml.add_representer(OrderedDict, self.represent_odict)
            yaml.dump(self.build_stairlight_config(), f)
        return template_file_name

    def create_mapping_file(
        self,
        unmapped: list[dict[str, Any]],
        prefix: str = MAPPING_CONFIG_PREFIX_DEFAULT,
    ) -> str:
        """Create a mapping template file

        Args:
            unmapped (list[dict]): Unmapped results
            prefix (str, optional):
                Prefix of the configuration file name.
                Defaults to MAPPING_CONFIG_PREFIX.

        Returns:
            str: Created file name
        """
        now = datetime.now(timezone.utc).strftime("%Y%m%d%H%M%S")
        template_file_name = f"{self.dir}/{prefix}_{now}.yaml"

        with open(template_file_name, "w") as f:
            yaml.add_representer(
                data_type=OrderedDict, representer=self.represent_odict
            )
            yaml.dump(
                self.build_mapping_config(unmapped_templates=unmapped),
                stream=f,
            )
        return template_file_name

    @staticmethod
    def represent_odict(
        dumper: yaml.Dumper, odict: OrderedDict
    ) -> yaml.nodes.MappingNode:
        """Set for dumping a YAML file in order of OrderedDict

        Args:
            dumper (yaml.Dumper): Dumper
            odict (OrderedDict): Ordered dict

        Returns:
            yaml.nodes.MappingNode: Mapping node
        """
        return dumper.represent_mapping(
            tag="tag:yaml.org,2002:map", mapping=odict.items()
        )

    @staticmethod
    def build_stairlight_config() -> OrderedDict:
        """Create a OrderedDict object for stairlight configurations

        Returns:
            OrderedDict: Template dict for stairlight.yaml
        """
        return OrderedDict(
            asdict(
                StairlightConfig(
                    Include=[
                        OrderedDict(asdict(StairlightConfigIncludeFile())),
                        OrderedDict(asdict(StairlightConfigIncludeGcs())),
                        OrderedDict(asdict(StairlightConfigIncludeRedash())),
                        OrderedDict(asdict(StairlightConfigIncludeDbt())),
                        OrderedDict(asdict(StairlightConfigIncludeS3())),
                    ],
                    Exclude=[OrderedDict(asdict(StairlightConfigExclude()))],
                    Settings=OrderedDict(asdict(StairlightConfigSettings())),
                )
            ),
        )

    def build_mapping_config(
        self, unmapped_templates: list[dict[str, Any]]
    ) -> OrderedDict:
        """Create a OrderedDict object for mapping configurations

        Args:
            unmapped_templates (list[dict[str, Any]]):
                Unmapped settings that Stairlight detects

        Returns:
            OrderedDict: Template dict for mapping.yaml
        """
        # Mapping section
        mappings: list[MappingConfigMapping] = []
        unmapped_template: dict[str, Any]
        for unmapped_template in unmapped_templates:
            # Tables.Parameters
            parameters: OrderedDict = OrderedDict()
            template: Template = unmapped_template[MapKey.TEMPLATE]

            if MapKey.PARAMETERS in unmapped_template:
                undefined_params: list[str] = unmapped_template.get(MapKey.PARAMETERS)
                for undefined_param in undefined_params:
                    splitted_params = undefined_param.split(".")
                    create_nested_dict(keys=splitted_params, results=parameters)

            # To avoid outputting empty fields
            if len(parameters) == 0:
                parameters = None

            mapping_table = MappingConfigMappingTable(
                TableName=get_default_table_name(template=template),
                Parameters=parameters,
            )
            table = OrderedDict(
                asdict(
                    mapping_table,
                    dict_factory=dict_factory,
                )
            )
            mapping: MappingConfigMapping = collect_mapping_attributes(
                template=template,
                tables=[table],
            )
            mappings.append(mapping)

        mapping_config = asdict(
            MappingConfig(
                Mapping=[OrderedDict(asdict(mapping)) for mapping in mappings],
            ),
            dict_factory=dict_factory,
        )
        return OrderedDict(mapping_config)


def dict_factory(d):
    return {k: v for (k, v) in d if v is not None}


def create_nested_dict(
    keys: list[str],
    results: dict[str, Any],
    density: int = 0,
    default_value: Any = None,
) -> None:
    """create nested dict from list

    Args:
        keys (list[str]): a list of dict keys
        results (dict[str, Any]): Nested dict
        density (int, optional): Density of the dict. Defaults to 0.
        default_value (any, optional): Default dict value. Defaults to None.
    """
    key = keys[density]
    if density < len(keys) - 1:
        if key not in results:
            results[key] = {}
        create_nested_dict(keys=keys, results=results[key], density=density + 1)
    else:
        results[key] = default_value



from __future__ import annotations

from dataclasses import dataclass, field
from typing import Any, OrderedDict

from src.stairlight.source.config import MappingConfigMapping, StairlightConfigInclude
from src.stairlight.source.template import TemplateSourceType as source_type


@dataclass
class StairlightConfigIncludeS3(StairlightConfigInclude):
    TemplateSourceType: str = source_type.S3.value
    ProjectId: str | None = None
    BucketName: str | None = None
    Regex: str | None = None
    DefaultTablePrefix: str | None = None


@dataclass
class MappingConfigMappingS3(MappingConfigMapping):
    TemplateSourceType: str = source_type.S3.value
    Uri: str | None = None
    Tables: list[OrderedDict[str, Any]] = field(default_factory=list)



from __future__ import annotations

import os
from dataclasses import asdict
from logging import getLogger
from typing import Any, Iterator

from sqlalchemy import create_engine, text
from sqlalchemy.engine.row import Row

from src.stairlight.source.config import (
    MappingConfig,
    MappingConfigMappingTable,
    StairlightConfig,
)
from src.stairlight.source.config_key import StairlightConfigKey
from src.stairlight.source.redash.config import StairlightConfigIncludeRedash
from src.stairlight.source.template import Template, TemplateSource, TemplateSourceType

logger = getLogger(__name__)


class RedashTemplate(Template):
    def __init__(
        self,
        mapping_config: MappingConfig,
        query_id: int,
        query_name: str,
        query_str: str = "",
        data_source_name: str = "",
    ):
        super().__init__(
            mapping_config=mapping_config,
            key=str(query_id),
            source_type=TemplateSourceType.REDASH,
        )
        self.query_id = query_id
        self.query_str = query_str
        self.uri = query_name
        self.data_source_name = data_source_name

    def find_mapped_table_attributes(self) -> Iterator[MappingConfigMappingTable]:
        """Get mapped tables as iterator

        Yields:
            Iterator[dict]: Mapped table attributes
        """
        mapping: Any
        for mapping in self._mapping_config.get_mapping():
            if mapping.TemplateSourceType != self.source_type.value:
                continue

            if (
                mapping.DataSourceName == self.data_source_name
                and mapping.QueryId == self.query_id
            ):
                for table_attributes in mapping.get_table():
                    yield table_attributes
                break

    def get_template_str(self) -> str:
        """Get template string that read from Redash
        Returns:
            str: Template string
        """
        return self.query_str

    def get_uri(self) -> str:
        """Get uri"""
        return super().get_uri()


class RedashTemplateSource(TemplateSource):
    REDASH_QUERIES = "sql/redash_queries.sql"
    WHERE_CLAUSE_TEMPLATES = {
        StairlightConfigKey.Redash.DATA_SOURCE_NAME: "data_sources.name = :data_source",
        StairlightConfigKey.Redash.QUERY_IDS: "queries.id IN :query_ids",
    }

    def __init__(
        self,
        stairlight_config: StairlightConfig,
        mapping_config: MappingConfig,
        include: StairlightConfigIncludeRedash,
    ) -> None:
        super().__init__(
            stairlight_config=stairlight_config,
            mapping_config=mapping_config,
        )
        self._include = include
        self.where_clause: list[str] = []
        self.conditions: dict[str, Any] = {}

    def search_templates(self) -> Iterator[Template]:
        """Search query template files

        Yields:
            Iterator[Template]: Attributes of query template files
        """
        results = self.get_redash_queries()
        for result in results:
            # see columns "src/stairlight/source/redash/sql/redash_queries.sql"
            yield RedashTemplate(
                mapping_config=self._mapping_config,
                query_id=result[0],
                query_name=result[1],
                query_str=result[2],
                data_source_name=result[3],
            )

    def get_redash_queries(self) -> list[Row]:
        """Get Redash queries

        Returns:
            list[Row]: Queries
        """
        current_dir = os.path.dirname(os.path.abspath(__file__))
        query_text = self.build_query_string(
            path=f"{current_dir}/{self.REDASH_QUERIES}"
        )
        data_source = self._include.DataSourceName
        query_ids = tuple(self._include.QueryIds) if self._include.QueryIds else None

        connection_str = self.get_connection_str()
        engine = create_engine(connection_str)
        with engine.connect() as conn:
            queries = conn.execute(  # type: ignore
                text(query_text),
                data_source=data_source,
                query_ids=query_ids,
            )
        return queries.fetchall()

    def build_query_string(self, path: str) -> str:
        """Build a query string

        Args:
            path (str): Path

        Returns:
            str: Query string
        """
        where_clauses: list[str] = []
        for key, value in self.WHERE_CLAUSE_TEMPLATES.items():
            if key in asdict(self._include).keys():
                where_clauses.append(value)

        base_query_string = self.read_query_string(path=path)
        return base_query_string + "WHERE " + " AND ".join(where_clauses)

    def read_query_string(self, path: str) -> str:
        """Read a query string

        Args:
            path (str): Path

        Returns:
            str: Query string
        """
        with open(path, "r", encoding="utf-8") as f:
            return f.read()

    def get_connection_str(self) -> str:
        """Get a database connection string

        Returns:
            str: Connection string
        """
        environment_variable_name = self._include.DatabaseUrlEnvironmentVariable
        connection_str = os.environ.get(environment_variable_name, "")
        if not connection_str:
            logger.error(f"{environment_variable_name} is not found.")
        return connection_str



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['stairlight',
 'stairlight.source',
 'stairlight.source.dbt',
 'stairlight.source.file',
 'stairlight.source.gcs',
 'stairlight.source.redash',
 'stairlight.source.s3']

package_data = \
{'': ['*'], 'stairlight.source.redash': ['sql/*']}

install_requires = \
['Jinja2>=2.10.3', 'PyYAML>=5.0']

extras_require = \
{':extra == "s3"': ['boto3-stubs[s3]>=1.24.17'],
 'dbt-bigquery': ['dbt-core>=1.4',
                  'dbt-bigquery>=1.1',
                  'google-cloud-bigquery>=2.0',
                  'protobuf>=3.20.2',
                  'networkx>=2.3,<2.8.1'],
 'gcs': ['google-cloud-storage>=2.0.0', 'protobuf>=3.20.2'],
 'redash': ['psycopg2>=2.9.3', 'SQLAlchemy>=1.4.31'],
 's3': ['boto3>=1.24.14']}

entry_points = \
{'console_scripts': ['stairlight = src.stairlight.cli:main']}

setup_kwargs = {
    'name': 'stairlight',
    'version': '0.7.1',
    'description': 'An end-to-end data lineage tool',
    'long_description': '<div align="center">\n  <img src="https://raw.githubusercontent.com/tosh2230/stairlight/main/img/stairlight_white.png" width="400" alt="Stairlight">\n</div>\n\n-----------------\n\n# Stairlight\n\n[![PyPi Version](https://img.shields.io/pypi/v/stairlight.svg?style=flat-square&logo=PyPi)](https://pypi.org/project/stairlight/)\n[![PyPi License](https://img.shields.io/pypi/l/stairlight.svg?style=flat-square)](https://pypi.org/project/stairlight/)\n[![PyPi Python Versions](https://img.shields.io/pypi/pyversions/stairlight.svg?style=flat-square)](https://pypi.org/project/stairlight/)\n[![Code style: black](https://img.shields.io/badge/code%20style-black-000000.svg?style=flat-square)](https://github.com/psf/black)\n[![CI](https://github.com/tosh2230/stairlight/actions/workflows/ci.yml/badge.svg)](https://github.com/tosh2230/stairlight/actions/workflows/ci.yml)\n\nAn end-to-end data lineage tool, detects table dependencies from SQL statements.\n\n<div align="left">\n  <img src="https://raw.githubusercontent.com/tosh2230/stairlight/main/img/drawio/concepts.drawio.png" width="1080" alt="concepts">\n</div>\n\n## Supported Data Sources\n\n| Data Source | Remarks |\n| --- | --- |\n| Local file system | Python Pathlib module |\n| [Amazon S3](https://aws.amazon.com/s3/) | Available for [Amazon Managed Workflows for Apache Airflow (MWAA)](https://aws.amazon.com/managed-workflows-for-apache-airflow/) |\n| [Google Cloud Storage](https://cloud.google.com/storage) | Available for [Google Cloud Composer](https://cloud.google.com/composer) |\n| [dbt](https://www.getdbt.com/) - [Google BigQuery](https://cloud.google.com/bigquery) | Using `dbt compile` command internally |\n| [Redash](https://redash.io/) | |\n\n## Installation\n\nThis package is distributed on [PyPI](https://pypi.org/project/stairlight/).\n\n```sh\n# The base package is for local file system only.\n$ pip install stairlight\n\n# Set extras when detecting from other data sources.\n# e.g. Amazon S3 and Google Cloud Storage\n$ pip install "stairlight[s3, gcs]"\n```\n\n| Data Source | TemplateSourceType | Extra |\n| --- | --- | --- |\n| Local file system | File | - |\n| Amazon S3 | S3 | s3 |\n| Google Cloud Storage | GCS | gcs |\n| dbt - Google Bigquery | dbt | dbt-bigquery |\n| Redash | Redash | redash |\n\n## Getting Started\n\nThere are 3 steps to use.\n\n```sh\n# 1: Initialize and set your data source settings\n$ stairlight init\n\n# 2: Map your SQL statements and tables\n$ stairlight map\n\n# 3: Get table dependencies\n$ stairlight\n```\n\n## Description\n\n### Input\n\n- SQL statements\n- Configuration YAML files\n    - stairlight.yaml: SQL statements locations and include/exclude conditions.\n    - mapping.yaml: For mapping SQL statements and tables.\n\n### Output\n\nStairlight outputs table dependencies as JSON format.\n\nTop-level keys are table names, and values represents tables that are the data source for each key\'s table.\n\n<details>\n\n<summary>Example</summary>\n\n```json\n{\n  "test_project.beam_streaming.taxirides_aggregation": {\n    "test_project.beam_streaming.taxirides_realtime": {\n      "TemplateSourceType": "File",\n      "Key": "tests/sql/main/union_same_table.sql",\n      "Uri": "/foo/bar/stairlight/tests/sql/main/union_same_table.sql",\n      "Lines": [\n        {\n          "LineNumber": 6,\n          "LineString": "    test_project.beam_streaming.taxirides_realtime"\n        },\n        {\n          "LineNumber": 15,\n          "LineString": "    test_project.beam_streaming.taxirides_realtime"\n        }\n      ]\n    }\n  },\n  "PROJECT_a.DATASET_b.TABLE_c": {\n    "PROJECT_A.DATASET_A.TABLE_A": {\n      "TemplateSourceType": "GCS",\n      "Key": "sql/one_line/one_line.sql",\n      "Uri": "gs://stairlight/sql/one_line/one_line.sql",\n      "Lines": [\n        {\n          "LineNumber": 1,\n          "LineString": "SELECT * FROM PROJECT_A.DATASET_A.TABLE_A WHERE 1 = 1"\n        }\n      ],\n      "BucketName": "stairlight",\n      "Labels": {\n        "Source": null,\n        "Test": "a"\n      }\n    }\n  },\n  "AggregateSales": {\n    "PROJECT_e.DATASET_e.TABLE_e": {\n      "TemplateSourceType": "Redash",\n      "Key": 5,\n      "Uri": "AggregateSales",\n      "Lines": [\n        {\n          "LineNumber": 1,\n          "LineString": "SELECT service, SUM(total_amount) FROM PROJECT_e.DATASET_e.TABLE_e GROUP BY service"\n        }\n      ],\n      "DataSourceName": "BigQuery",\n      "Labels": {\n        "Category": "Sales"\n      }\n    }\n  },\n  "dummy.dummy.example_b": {\n    "PROJECT_t.DATASET_t.TABLE_t": {\n      "TemplateSourceType": "dbt",\n      "Key": "tests/dbt/project_01/target/compiled/project_01/models/b/example_b.sql",\n      "Uri": "/foo/bar/stairlight/tests/dbt/project_01/target/compiled/project_01/models/b/example_b.sql",\n      "Lines": [\n        {\n          "LineNumber": 1,\n          "LineString": "select * from PROJECT_t.DATASET_t.TABLE_t where value_a = 0 and value_b = 0"\n        }\n      ]\n    }\n  },\n  "PROJECT_as.DATASET_bs.TABLE_cs": {\n    "PROJECT_A.DATASET_A.TABLE_A": {\n      "TemplateSourceType": "S3",\n      "Key": "sql/one_line/one_line.sql",\n      "Uri": "s3://stairlight/sql/one_line/one_line.sql",\n      "Lines": [\n        {\n          "LineNumber": 1,\n          "LineString": "SELECT * FROM PROJECT_A.DATASET_A.TABLE_A WHERE 1 = 1"\n        }\n      ],\n      "BucketName": "stairlight",\n      "Labels": {\n        "Source": null,\n        "Test": "a"\n      }\n    }\n  }\n}\n```\n\n</details>\n\n### Collecting patterns\n\n#### Centralization\n\n<div align="left">\n  <img src="https://raw.githubusercontent.com/tosh2230/stairlight/main/img/drawio/centralization.drawio.png" width="800" alt="centralization">\n</div>\n\n#### Agents\n\n<div align="left">\n  <img src="https://raw.githubusercontent.com/tosh2230/stairlight/main/img/drawio/agents.drawio.png" width="800" alt="agents">\n</div>\n\n## Configuration\n\nExamples can be found [here](https://github.com/tosh2230/stairlight/tree/main/tests/config), used for unit testing in CI.\n\n### stairlight.yaml\n\n\'stairlight.yaml\' is for setting up Stairlight itself. It is responsible for specifying SQL statements to be read.\n\n`stairlight init` creates a template of stairlight.yaml.\n\n<details>\n\n<summary>Example</summary>\n\n```yaml\nInclude:\n  - TemplateSourceType: File\n    FileSystemPath: "./tests/sql"\n    Regex: ".*/*.sql$"\n    DefaultTablePrefix: "PROJECT_A"\n  - TemplateSourceType: GCS\n    ProjectId: null\n    BucketName: stairlight\n    Regex: "^sql/.*/*.sql$"\n    DefaultTablePrefix: "PROJECT_A"\n  - TemplateSourceType: Redash\n    DatabaseUrlEnvironmentVariable: REDASH_DATABASE_URL\n    DataSourceName: BigQuery\n    QueryIds:\n      - 1\n      - 3\n      - 5\n  - TemplateSourceType: dbt\n    ProjectDir: tests/dbt/project_01\n    ProfilesDir: tests/dbt\n    Vars:\n      key_a: value_a\n      key_b: value_b\n  - TemplateSourceType: S3\n    BucketName: stairlight\n    Regex: "^sql/.*/*.sql$"\n    DefaultTablePrefix: "PROJECT_A"\nExclude:\n  - TemplateSourceType: File\n    Regex: "main/exclude.sql$"\nSettings:\n  MappingPrefix: "mapping"\n```\n\n</details>\n\n### mapping.yaml\n\n\'mapping.yaml\' is used to define relationships between input SELECT statements and tables.\n\n`stairlight map` creates a template of mapping.yaml and attempts to read from data sources specified in stairlight.yaml.\nIf successfully read, it outputs settings that have not yet configured in an existing \'mapping.yaml\' file.\n\n<details>\n\n<summary>Example</summary>\n\n```yaml\nGlobal:\n  Parameters:\n    DESTINATION_PROJECT: stairlight\n    params:\n      PROJECT: 1234567890\n      DATASET: public\n      TABLE: taxirides\nMapping:\n  - TemplateSourceType: File\n    FileSuffix: "tests/sql/main/union_same_table.sql"\n    Tables:\n      - TableName: "test_project.beam_streaming.taxirides_aggregation"\n        Parameters:\n          params:\n            source_table: source\n            destination_table: destination\n        IgnoreParameters:\n          - execution_date.add(days=1).isoformat()\n  - TemplateSourceType: GCS\n    Uri: "gs://stairlight/sql/one_line/one_line.sql"\n    Tables:\n      - TableName: "PROJECT_a.DATASET_b.TABLE_c"\n  - TemplateSourceType: Redash\n    QueryId: 5\n    DataSourceName: metadata\n    Tables:\n      - TableName: New Query\n        Parameters:\n          table: dashboards\n        Labels:\n          Category: Redash test\n  - TemplateSourceType: dbt\n    ProjectName: project_01\n    FileSuffix: tests/dbt/project_01/target/compiled/project_01/models/example/my_first_dbt_model.sql\n    Tables:\n      - TableName: dummy.dummy.my_first_dbt_model\n  - TemplateSourceType: S3\n    Uri: "s3://stairlight/sql/one_line/one_line.sql"\n    Tables:\n      - TableName: "PROJECT_as.DATASET_bs.TABLE_cs"\nExtraLabels:\n  - TableName: "PROJECT_A.DATASET_A.TABLE_A"\n    Labels:\n      Source: Null\n      Test: a\n```\n\n</details>\n\n#### Global Section\n\nThis section is for global configurations.\n\n`Parameters` is used to set common parameters. If conflicts has occurred with `Parameters` in mapping section, mapping section\'s parameters will be used in preference to global.\n\n#### Mapping Section\n\nMapping section is used to define relationships between input SELECT statements and tables that created as a result of query execution.\n\n`Parameters` allows you to reflect settings in [jinja](https://jinja.palletsprojects.com/) template variables embedded in statements. If multiple settings are applied to a statement using jinja template, the statement will be read as if there were the same number of queries as the number of settings.\n\nIn contrast, `IgnoreParameters` handles a list to ignore when rendering queries.\n\n#### Extra labels Section\n\nThis section sets labels to tables that appears only in queries.\n\n## Arguments and Options\n\n```txt\n$ stairlight --help\nusage: stairlight [-h] [-c CONFIG] [--save SAVE] [--load LOAD] {init,check,up,down} ...\n\nAn end-to-end data lineage tool, detects table dependencies by SQL SELECT statements.\nWithout positional arguments, return a table dependency map as JSON format.\n\npositional arguments:\n  {init,map,check,list,up,down}\n    init                create a new Stairlight configuration file\n    map (check)         create a new configuration file about undefined mappings\n    list                return all ( tables | URIs )\n    up                  return upstairs ( tables | URIs )\n    down                return downstairs ( tables | URIs )\n\noptional arguments:\n  -h, --help            show this help message and exit\n  -c CONFIG, --config CONFIG\n                        set a Stairlight configuration directory\n  -q, --quiet           keep silence\n  --save SAVE           A file path where map results will be saved.\n                        You can choose from local file system, GCS, S3.\n  --load LOAD           A file path where map results are saved.\n                        You can choose from local file system, GCS, S3.\n                        It can be specified multiple times.\n```\n\n### init\n\n`stairlight init` creates a new Stairlight configuration file.\n\n```txt\n$ stairlight init --help\nusage: stairlight init [-h] [-c CONFIG]\n\noptional arguments:\n  -h, --help            show this help message and exit\n  -c CONFIG, --config CONFIG\n                        set a Stairlight configuration directory\n  -q, --quiet           keep silence\n```\n\n### map(check)\n\n`stairlight map` creates a new configuration file about undefined settings. `stairlight check` is an alias.\nOptions are the same as `stairlight init`.\n\n### list\n\n`stairlight list` outputs all of tables or SQL URIs.\n\n- Output option(`-o`, `--output`) determines the output type, tables or URIs.\n\n### up\n\n`stairlight up` outputs tables or SQL URIs located upstream(upstairs) from the specified table.\n\n- Use table(`-t`, `--table`) or label(`-l`, `--label`) option to specify tables to search.\n- Output option(`-o`, `--output`) is same as `stairlight list`.\n- Recursive option(`-r`, `--recursive`) is set, Stairlight will find dependencies recursively and output as a list.\n- Verbose option(`-v`, `--verbose`) is set, Stairlight will add detailed information and output it as a dict.\n\n```txt\n$ stairlight up --help\nusage: stairlight up [-h] [-c CONFIG] [--save SAVE] [--load LOAD] (-t TABLE | -l LABEL) [-o {table,uri}]\n                     [-v] [-r]\n\noptional arguments:\n  -h, --help            show this help message and exit\n  -c CONFIG, --config CONFIG\n                        set a Stairlight configuration directory\n  -q, --quiet           keep silence\n  --save SAVE           A file path where mapped results will be saved.\n                        You can choose from local file system, GCS, S3.\n  --load LOAD           A file path where mapped results are saved.\n                        You can choose from local file system, GCS, S3.\n                        It can be specified multiple times.\n  -t TABLE, --table TABLE\n                        table names that Stairlight searches for, can be specified\n                        multiple times. e.g. -t PROJECT_a.DATASET_b.TABLE_c -t\n                        PROJECT_d.DATASET_e.TABLE_f\n  -l LABEL, --label LABEL\n                        labels set for the table in mapping configuration, can be specified multiple times.\n                        The separator between key and value should be a colon(:).\n                        e.g. -l key_1:value_1 -l key_2:value_2\n  -o {table,uri}, --output {table,uri}\n                        output type\n  -v, --verbose         return verbose results\n  -r, --recursive       search recursively\n```\n\n### down\n\n`stairlight down` outputs tables or SQL URIs located downstream(downstairs) from the specified table.\nOptions are the same as `stairlight up`.\n\n## Use as a library\n\nStairlight can also be used as a library.\n\n[tosh2230/stairlight-app](https://github.com/tosh2230/stairlight-app) is a sample web application rendering table dependency graph with Stairlight, using Graphviz, Streamlit and Google Cloud Run.\n',
    'author': 'tosh2230',
    'author_email': 'rev.to12@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.7.2',
}


setup(**setup_kwargs)



import logging
from abc import ABC
from typing import Any

logger = logging.getLogger()


GCS_URI_SCHEME = "gs://"
S3_URI_SCHEME = "s3://"


class Key(ABC):
    def __init__(self) -> None:
        super().__init__()

    def __setattr__(self, name: str, value: Any) -> None:
        if name in self.__dict__:
            raise TypeError()
        else:
            self.__setattr__(name=name, value=value)


class StairlightConfigKey(Key):
    INCLUDE_SECTION = "Include"
    EXCLUDE_SECTION = "Exclude"
    SETTING_SECTION = "Settings"

    TEMPLATE_SOURCE_TYPE = "TemplateSourceType"
    DEFAULT_TABLE_PREFIX = "DefaultTablePrefix"
    REGEX = "Regex"

    MAPPING_PREFIX = "MappingPrefix"

    class File(Key):
        FILE_SYSTEM_PATH = "FileSystemPath"

    class Gcs(Key):
        PROJECT_ID = "ProjectId"
        BUCKET_NAME = "BucketName"

    class Redash(Key):
        DATABASE_URL_ENV_VAR = "DatabaseUrlEnvironmentVariable"
        DATA_SOURCE_NAME = "DataSourceName"
        QUERY_IDS = "QueryIds"

    class Dbt(Key):
        PROJECT_DIR = "ProjectDir"
        PROFILES_DIR = "ProfilesDir"
        TARGET = "Target"
        VARS = "Vars"

    class S3(Key):
        BUCKET_NAME = "BucketName"


class MappingConfigKey(Key):
    GLOBAL_SECTION = "Global"
    MAPPING_SECTION = "Mapping"

    METADATA_SECTION = "Metadata"  # Deprecated
    EXTRA_LABELS_SECTION = "ExtraLabels"

    TEMPLATE_SOURCE_TYPE = "TemplateSourceType"
    TABLES = "Tables"
    TABLE_NAME = "TableName"
    IGNORE_PARAMETERS = "IgnoreParameters"
    PARAMETERS = "Parameters"
    LABELS = "Labels"

    class File(Key):
        FILE_SUFFIX = "FileSuffix"

    class Gcs(Key):
        URI = "Uri"
        BUCKET_NAME = "BucketName"

    class Redash(Key):
        QUERY_ID = "QueryId"
        DATA_SOURCE_NAME = "DataSourceName"

    class Dbt(Key):
        FILE_SUFFIX = "FileSuffix"
        PROJECT_NAME = "ProjectName"

    class S3(Key):
        URI = "Uri"
        BUCKET_NAME = "BucketName"


class DbtProjectKey(Key):
    PROJECT_NAME = "name"
    MODEL_PATHS = "model-paths"
    TARGET_PATH = "target-path"
    PROFILE = "Profile"


class MapKey(Key):
    TABLE_NAME = "TableName"
    TEMPLATE_SOURCE_TYPE = "TemplateSourceType"
    KEY = "Key"
    URI = "Uri"
    LINES = "Lines"
    LINE_NUMBER = "LineNumber"
    LINE_STRING = "LineString"
    BUCKET_NAME = "BucketName"
    LABELS = "Labels"
    DATA_SOURCE_NAME = "DataSourceName"

    TEMPLATE = "Template"
    PARAMETERS = "Parameters"
